@@ -83,18 +83,22 @@ _vzctl()
 		--dgramrcvbuf --oomguarpages --lockedpages --privvmpages \
 		--shmpages --numfile --numflock --numpty --numsiginfo \
 		--dcachesize --numiptent --physpages --swappages \
-		--cpuunits --cpulimit --cpus --cpumask \
+		--vm_overcommit --cpuunits --cpulimit --cpus --cpumask \
 		--iptables --netdev_add --netdev_del --netif_add --netif_del \
 		--diskquota --mount_opts \
 		--diskspace --diskinodes --quotatime --quotaugidlimit \
 		--capability --devnodes --devices --applyconfig \
-		--name --ioprio --setmode --meminfo --bootorder --features \
-		--reset_ub"
-	local vzctl_start_opts="--wait --force"
+		--name --ioprio --iolimit --iopslimit \
+		--setmode --meminfo --bootorder --features \
+		--reset_ub --stop-timeout"
+	local vzctl_start_opts="--wait --force --skip-fsck"
 	local vzctl_stop_opts="--fast --skip-umount"
 	local vzctl_restart_opts="$vzctl_start_opts $vzctl_stop_opts"
 	local vzctl_snapshot_create_opts="--id --name --description \
 		--skip-suspend --skip-config"
+	local vzctl_snapshot_mount_opts="--id --target"
+	local vzctl_snapshot_list_opts="-H --no-header -o --output --id \
+		-L --list --help"
 
 	local iptables_names="iptable_filter iptable_mangle ipt_limit
 		ipt_multiport ipt_tos ipt_TOS ipt_REJECT ipt_TCPMSS
@@ -113,6 +117,7 @@ _vzctl()
 
 	if test -f /proc/vz/ploop_minor; then
 		vzctl_cmds+=" convert compact snapshot \
+			snapshot-mount snapshot-umount \
 			snapshot-switch snapshot-delete snapshot-list"
 	fi
 
@@ -147,14 +152,9 @@ _vzctl()
 			# flag or option
 			case $prev in
 			--ostemplate)
-				# get the template path from the vz.conf
-				local vztmpl=`grep ^TEMPLATE /etc/vz/vz.conf | cut -d "=" -f 2`
-				COMPREPLY=( $( compgen -W "$(command ls $vztmpl/cache/*.tar{,.gz,.xz,.bz2} 2>/dev/null |
-						    sed -e "s#^$vztmpl/cache/##" -e 's#\.tar\.*[gbxz2]*$##';
-						    /usr/libexec/vzctl/scripts/vps-download --list)" -- $cur ) )
-
+				COMPREPLY=( $( compgen -W "$(/usr/sbin/vztmpl-dl --list-all)" -- $cur ) )
 				;;
-			--onboot|--disabled|--noatime|--diskquota)
+			--onboot|--disabled|--diskquota)
 				COMPREPLY=( $( compgen -W "yes no" -- $cur ) )
 				;;
 			--setmode)
@@ -257,6 +257,37 @@ _vzctl()
 				# makes no sense for other parameters (UBCs
 				# etc). So no action for now.
 				;;
+			--id|--uuid)
+				# Only allow this completion
+				# when user has root privileges
+				[ "$EUID" -ne 0 ] && exit 1
+				# Get command and CTID
+				local cmd=${COMP_WORDS[1]}
+				local ve=${COMP_WORDS[2]}
+				if [ ${cmd::2} = "--" ] ; then
+					# --verbose or --quiet used
+					cmd=${COMP_WORDS[2]}
+					ve=${COMP_WORDS[3]}
+				fi
+				local ids
+				case ${cmd} in
+				    snapshot-umount)
+					# For umount, only give mounted IDs
+					ids="`/usr/sbin/vzctl snapshot-list $ve -H -ouuid,device 2>/dev/null | awk 'NF==2 {print $1}' | tr -d '{}'`"
+					;;
+				    snapshot-*)
+					# For other commands, give all IDs
+					ids="`/usr/sbin/vzctl snapshot-list $ve -H -ouuid 2>/dev/null | tr -d '{}'`"
+					;;
+				    *)
+					# Do not give anything for other cmds
+					# including 'snapshot' itself
+					# since it requires a new unused ID
+					exit 0
+					;;
+				esac
+				COMPREPLY=( $( compgen -W "$ids" -- $cur ) )
+				;;
 			*)
 				if [[ "${prev::2}" != "--" || "$prev" = "--save" ]]; then
 					# List options
@@ -291,9 +322,15 @@ _vzctl()
 					snapshot)
 						COMPREPLY=( $( compgen -W "$vzctl_snapshot_create_opts" -- $cur ) )
 						;;
-					snapshot-switch|snapshot-delete)
+					snapshot-switch|snapshot-delete|snapshot-umount)
 						COMPREPLY=( $( compgen -W "--id" -- $cur ) )
 						;;
+					snapshot-mount)
+						COMPREPLY=( $( compgen -W "$vzctl_snapshot_mount_opts" -- $cur ) )
+						;;
+					snapshot-list)
+						COMPREPLY=( $( compgen -W "$vzctl_snapshot_list_opts" -- $cur ) )
+						;;
 					*)
 						;;
 					esac
